/*******************************************************************************
 * Copyright (c) 2008 Intel Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Rodolfo Kohn - initial API and implementation
 *    Marcelo Lorenzatti, Miriam Estela - Management of multiple configurations
      and filters .
 *******************************************************************************
 Product name: Plugin Extension for CIM  
 Creation date: 1-Nov-2007
 Description: 
 This file handles  subscriptions to the related events.

 Modification History
 02/25/08        Marcelo Lorenzati      Improving the appearance of the code in 
                 order to align to sblim coding rules.
                 Added some error cases.
 09/26/08        Marcelo Lorenzati      Added a Configuration Management module
 10/01/08        Marcelo Lorenzati      Removing of debug output and comments
********************************************************************************/

#include <stdio.h>
#include <pthread.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <signal.h>

#include <cmci.h>
#include <native.h>

#include "libpec_configurationmanagement.h"

#define LISTEN_PORT 2880
#define BACKLOG 10
#define ADDRLEN INET_ADDRSTRLEN
#define BUFF_SIZE 10000
#define BUFFER_SIZE 256
#define QUERY_BUFFER_SIZE 10000
#define EVENTID_SIZE 20
#define DEST_PORT 5050
#define ADDRLEN INET_ADDRSTRLEN
#define IND_ID "<PROPERTY NAME=\"IndicationIdentifier\" TYPE=\"string\">"
#define PERCEIVED_SEV "<PROPERTY NAME=\"PerceivedSeverity\" TYPE=\"uint16\">"
#define PROB_CAUSE_DESC "<PROPERTY NAME=\"ProbableCauseDescription\" TYPE=\"string\">"
#define IND_TIME "<PROPERTY NAME=\"IndicationTime\" TYPE=\"datetime\">"
#define CLASS_NAME "<PROPERTY NAME=\"ClassName\" TYPE=\"string\">"
#define CL_KEY "Content-Length: "
#define TEST_ITERATION_MAX 1200

#define DEFAULT_CIM_HOST "localhost"
#define DEFAULT_USER "root"
#define DEFAULT_PASSWORD "pec123"

char gDestaddr[ADDRLEN + 1];
int listenEnd = 0;
static int portidx=0;
static int getValue(const char *buf, const char *propertyTag, char *value, int len);

static int passMessage(char *buff, int len);

static int sendMsg(const char *buf);

void sigEnd(int signo);

int listenToIndications();

int generateQuery(char *query, int len, char *eventID);

static int getValue(const char *buf, const char *propertyTag, char *value, int len)
{
    const char *ptr;
    const char *ptr2;
    
    ptr = strstr(buf, propertyTag);
    if (ptr)
    {
        ptr += strlen(propertyTag);
        ptr += (strlen("<VALUE>") + 1);
        ptr2 = strstr(ptr, "</VALUE>");
        if (ptr2)
        {
            memcpy(value, ptr, ((ptr2 - ptr) > (len - 1)) ? (len - 1) : (ptr2 - ptr));
            value[((ptr2 - ptr) > (len - 1)) ? (len - 1): (ptr2 - ptr)] = 0;
            return 0;
        }
    }
    return 1;
}


static int passMessage(char *buff, int len)
{
    static char value[512];
    static char result[1024];
    
    printf("---------------------------------------------------------------------\n");   

    getValue(buff, IND_ID, value, 512);
    printf("INDICATION IDENTIFIER: %s\n", value);
    strcpy(result, value);
    strcat(result, " - ");

    getValue(buff, CLASS_NAME, value, 512);
    printf("INDICATION TYPE: %s\n", value);
    strcat(result, value);
    strcat(result, " - ");

    getValue(buff, PERCEIVED_SEV, value, 512);
    printf("INDICATION SEVERITY: %s\n", value);
    strcat(result, value);
    strcat(result, " - ");

    getValue(buff, PROB_CAUSE_DESC, value, 512);
    printf("INDICATION DESCRIPTION: %s\n", value);
    strcat(result, value);
    strcat(result, " - ");

    getValue(buff, IND_TIME, value, 512);
    printf("INDICATION TIMESTAMP: %s\n", value);
    strcat(result, value);
   
    printf("---------------------------------------------------------------------\n");
    
    return 0;        
}

void sigEnd(int signo)
{
    printf("Finalization signal received - Wait for termination\n");
    listenEnd = 1;
    return;
}


int listenToIndications()
{
    int sock;
    int res;
    struct sockaddr_in myAddr;
    struct sockaddr_in clientaddr;
    int clisize = sizeof(clientaddr);
    int clisock;
    char addrBuffer[ADDRLEN + 1];
    char buff[BUFF_SIZE];
    int bytesRead;
    int i;
    char *sepPtr;
    char *contLenPtr;
    char *contLenEndPtr;
    char contLengthStr[10];
    long contLength;
    int totalBytesRead;

    printf("entered listenToIndications\n");
    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock == -1)
    {
        printf("Error opening Packet Socket: %d\n", errno);
        return -1;
    }

    bzero(&myAddr, sizeof(myAddr));
    myAddr.sin_family = AF_INET;
    myAddr.sin_port = htons(LISTEN_PORT + portidx);
    myAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    res = bind(sock, (struct sockaddr *)&myAddr, sizeof(struct sockaddr_in));
    if (res == -1)
    {
        printf("Error binding Socket: %d\n", errno);
        return -2;
    }
    res = listen(sock, BACKLOG);
    if (res == -1)
    {
        printf("Error listening to Socket: %d\n", errno);
        return -3;
    }
    //while(!listenEnd)
    {
        clisock = accept(sock, (struct sockaddr *)&clientaddr, &clisize);
        if (clisock == -1)
        {
            if (errno != EINTR)
            {
                printf("Error accepting connection: %d\n", errno);
                return -4;
            }
        }
        else
        {
            if (inet_ntop(AF_INET, (const void *)&clientaddr.sin_addr, addrBuffer, ADDRLEN) == NULL)
            {
                printf("Error converting client address: %d\n", errno);
                return -5;
            }
            printf("Accepted connection @ %s : %d\n", addrBuffer, ntohs(clientaddr.sin_port));
            if ((bytesRead = read(clisock, (void *)buff, BUFF_SIZE)) == -1)
            {
                if (errno != EINTR)
                {
                    printf("Error reading from client socket: %d\n", errno);
                    return -6;
                }
            }

            sepPtr = strstr(buff, "\r\n\r\n");
            if (sepPtr)
            {
                contLenPtr = strstr(buff, CL_KEY);
                contLenEndPtr = strstr(contLenPtr, "\r\n");
                strncpy(contLengthStr, (contLenPtr + strlen(CL_KEY)), (contLenEndPtr - (contLenPtr + strlen(CL_KEY))));
                contLengthStr[(contLenEndPtr - (contLenPtr + strlen(CL_KEY)))] = 0;
                contLength = atol(contLengthStr);
                if (bytesRead < (((sepPtr + 4) - buff) + contLength))
                {
                    totalBytesRead = bytesRead;
                    while((totalBytesRead < (((sepPtr + 4) - buff) + contLength)) && !listenEnd)
                    {
                        if ((bytesRead = read(clisock, (void *)(buff + totalBytesRead), (BUFF_SIZE - totalBytesRead))) == -1)
                        {
                            if (errno != EINTR)
                            {
                                printf("Error reading from client socket: %d\n", errno);
                                return -6;
                            }
                        }
                        totalBytesRead += bytesRead;
                        printf("Message received: \n\n");
                    }
                    if (!listenEnd)
                    {
                        for(i = 0; i < totalBytesRead; i++)
                        {
                            printf("%c", buff[i]);
                        }
                        printf("\n\n");
                        passMessage(buff, totalBytesRead);
                    }
                }
            }

            if (!listenEnd)
            {
                sprintf(buff, 
                    "HTTP/1.1 200 Ok\n\r"
                    "Date: Tue, 30 Nov 1999 00:22:34 GMT\n\r"
                    "Cache-Control: no-cache\n\r"
                    "Server: sfcb/1.2.4 (CIMOM)\n\r"
                    "Content-Type: application/xml; charset=\"utf-8\"\n\r"
                    "Man: http://www.dmtf.org/cim/mapping/http/v1.0 ; ns=02\n\r"
                    "02-CIMOperation: MethodResponse\n\r"
                    "Content-Length: 209\n\r\n\r"
                    "<?xml version=\"10\" ?><CIM CIMVERSION=\"2.0\" DTDVERSION=\"2.0\"><MESSAGE ID=\"1\" PROTOCOLVERSION=\"1.0\"><SIMPLEEXPRSP><EXPMETHODRESPONSE NAME=\"ExportIndication\"></EXPMETHODRESPONSE></SIMPLEEXPRSP></MESSAGE></CIM>\n\r");
                write(clisock, (void *) buff, strlen(buff));
            }
            close(clisock);
        }
    }
    if (close(sock))
    {
        printf("Error closing socket\n");
    }
    return 0;
}


int generateQuery(char *query, int len, char *eventID)
{
    sprintf(query, "SELECT * FROM CIM_AlertIndication WHERE IndicationIdentifier='0xFFEF0091'");//WHERE IndicationIdentifier='0xFFEF0091' AND SourceClass='CIM_Process' OR IndicationIdentifier='0xFFEF1174'
    return 0;
}

/*This function connects to the sfcb broker and creates a Handler a filter and subscriber to events of the plugins*/
int main(int argc, char **argv)
{
    CMCIClient *cc;
    CMPIObjectPath * objectpathf, *objectpathf_r;
    CMPIObjectPath * objectpathh, *objectpathh_r;
    CMPIObjectPath * objectpaths, *objectpaths_r;
    CMPIInstance * instancef;
    CMPIInstance * instanceh;
    CMPIInstance * instances;
    CMPIStatus status;
    char hostName[512];
    char *cim_host, *cim_host_passwd, *cim_host_userid;
    CMPIValue value;
    char buffer[BUFFER_SIZE];
    char query[QUERY_BUFFER_SIZE];
    char eventID[EVENTID_SIZE];
    struct sigaction provAction, oldAction;
    int testcount;

    if (argc < 2)
    {
        strncpy(gDestaddr, "127.0.0.1", ADDRLEN);
    }
    else
    {
        strncpy(gDestaddr, argv[1], ADDRLEN);
    }
    gDestaddr[ADDRLEN] = 0;
    provAction.sa_handler = sigEnd;
    sigemptyset(&provAction.sa_mask);
    provAction.sa_flags = 0;
    if (sigaction(SIGINT, &provAction, &oldAction) != 0)
    {
        printf("Error setting provider SIGCHLD signal handler\n");
    }

    pecConfigurationManagementInit();
     
    cim_host = pecConfigurationManagementGetValue("CIM_SERVER");
    if (cim_host == NULL) { cim_host = DEFAULT_CIM_HOST; }
    cim_host_userid = pecConfigurationManagementGetValue("USER");
    if (cim_host_userid == NULL) { cim_host_userid = DEFAULT_USER; }
    cim_host_passwd = pecConfigurationManagementGetValue("PASSWORD");
    if (cim_host_passwd == NULL) { cim_host_passwd = DEFAULT_PASSWORD; }

    printf("--------------------Test: Iterating %d times...--------------------\n", TEST_ITERATION_MAX);
    for ( testcount = 0; testcount < TEST_ITERATION_MAX; testcount++ )
    {
        printf("----------------------Test: Init Iteration %d...----------------------\n", testcount);
        cc = cmciConnect(cim_host, NULL, "5988", cim_host_userid, cim_host_passwd, NULL);
    
        printf("\n---------------------------------------------------------------------\n");
        printf("Testing createInstance() ...\n");
        objectpathh = newCMPIObjectPath("root/interop", "CIM_IndicationHandlerCIMXML", NULL);
        CMAddKey(objectpathh, "SystemCreationClassName", "CIM_System", CMPI_chars);
        CMAddKey(objectpathh, "SystemName", "127.0.0.1", CMPI_chars);
        CMAddKey(objectpathh, "CreationClassName", "CIM_IndicationHandlerCIMXML", CMPI_chars);
        CMAddKey(objectpathh, "Name", "pecclient", CMPI_chars);
        instanceh = newCMPIInstance(objectpathh, NULL);
        portidx++;
        sprintf(buffer, "http://127.0.0.1:%d", LISTEN_PORT + portidx);
        CMSetProperty(instanceh, "Destination", buffer, CMPI_chars);
        value.uint16 = 2;
        CMSetProperty(instanceh, "PersistenceType", &value, CMPI_uint16);
        objectpathh_r = cc->ft->createInstance(cc, objectpathh, instanceh, &status);
        printf("createInstance() rc=%d, msg=%s\n", status.rc, 
               (status.msg)? (char *)status.msg->hdl : NULL);
        if (status.msg) CMRelease(status.msg);
        if (status.rc) return 1;
    
        objectpathf = newCMPIObjectPath("root/interop", "CIM_IndicationFilter", NULL);
        CMAddKey(objectpathf, "SystemCreationClassName", "CIM_System", CMPI_chars);
        CMAddKey(objectpathf, "SystemName", "127.0.0.1", CMPI_chars);
        CMAddKey(objectpathf, "CreationClassName", "CIM_IndicationFilter", CMPI_chars);
        CMAddKey(objectpathf, "Name", "pecclient", CMPI_chars);
        instancef = newCMPIInstance(objectpathf, NULL);
        CMSetProperty(instancef, "SourceNamespace", "root/interop", CMPI_chars);
        sprintf(eventID, "4351");
        generateQuery(query, QUERY_BUFFER_SIZE, eventID);
        CMSetProperty(instancef, "Query", query, CMPI_chars);
        CMSetProperty(instancef, "QueryLanguage", "WQL", CMPI_chars);
        objectpathf_r = cc->ft->createInstance(cc, objectpathf, instancef, &status);
        printf("createInstance 2() rc=%d, msg=%s\n", status.rc,
               (status.msg)? (char *)status.msg->hdl : NULL);
        if (status.msg) CMRelease(status.msg);
        if (status.rc) return 1;

        objectpaths = newCMPIObjectPath("root/interop", "CIM_IndicationSubscription", NULL);
        value.ref = objectpathf;
        CMAddKey(objectpaths, "Filter", &value, CMPI_ref);
        value.ref = objectpathh;
        CMAddKey(objectpaths, "Handler", &value, CMPI_ref);
        instances = newCMPIInstance(objectpaths, NULL);
        value.uint16 = 2;
        CMSetProperty(instances, "SubscriptionState", &value, CMPI_uint16); 
        value.uint16 = 2;
        CMSetProperty(instances, "OnFatalErrorPolicy", &value, CMPI_uint16); 
        value.uint16 = 2;
        CMSetProperty(instances, "RepeatNotificationPolicy", &value, CMPI_uint16); 
        objectpaths_r = cc->ft->createInstance(cc, objectpaths, instances, &status);
        /* Print the results */
        printf("createInstance 3() rc=%d, msg=%s\n", status.rc,
               (status.msg)? (char *)status.msg->hdl : NULL);
        if (status.msg) CMRelease(status.msg);
        if (status.rc) return 1;
    
        if (instanceh) CMRelease(instanceh);
        if (instancef) CMRelease(instancef);
        if (instances) CMRelease(instances);
        listenToIndications();
        status = cc->ft->deleteInstance(cc, objectpaths);
        if (status.rc) 
        {
            printf("error deleting Instance subscription\n");
        }
        status = cc->ft->deleteInstance(cc, objectpathh);
        if (status.rc)
        {
            printf("error deleting Instance handler\n");
        }
        status = cc->ft->deleteInstance(cc, objectpathf);
        if (status.rc)
        {
            printf("error deleting Instance filter\n");
        }
        if (objectpathh) CMRelease(objectpathh);
        if (objectpathh_r) CMRelease(objectpathh_r);
        if (objectpathf) CMRelease(objectpathf);
        if (objectpathf_r) CMRelease(objectpathf_r);
        if (objectpaths) CMRelease(objectpaths);
        if (objectpaths_r) CMRelease(objectpaths_r);
        if (status.msg) CMRelease(status.msg);
        if (cc) CMRelease(cc);
        printf("\n----------------------Test: Close Iteration %d...---------------------\n\n", testcount);
        if (listenEnd)
        {
            break;
        }
        else
        {
            sleep(1);
        } 
    }
    if (sigaction(SIGINT, &oldAction, NULL) != 0)
    {
        printf("Error setting old SIGCHLD signal handler\n");
    }
    return 0;
}
