/*******************************************************************************
 * Copyright (c) 2008 Intel Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Rodolfo Kohn - initial API and implementation
 *    Marcelo Lorenzatti, Miriam Estela - Management of multiple configurations
      and filters .
 *******************************************************************************
 Product name: Plugin Extension for CIM  
 Creation date: 1-Nov-2007
 Description: 
 This file is the Indication Provider that reads the plug-in configuration
 And executes each plug-in in a different thread.
 
 Modification History
 02/19/2008      Marcelo Lorenzati, Miriam Estela    Fixed to read more than one
                 functionality per plugin
********************************************************************************/

#define CMPI_PLATFORM_LINUX_GENERIC_GNU

#include <stdio.h>
#include <pthread.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sched.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include "cmpidt.h"
#include "cmpift.h"
#include "cmpimacs.h"

#include "pec_plugins.h"
#include "pec_plugin_indication.h"

#define WHERE_KEY "WHERE"
#define MAX_ENV_STR 1
#define ARRAY_SIZE 20
#define ARGS_NUM 20
#define MAX_SENTENCE_LEN 48
#define MAX_READ_SIZE 1024
#define MAX_STR_LEN 1024
#define MAX_BUF_LEN 10000
#define MAX_FILTER_LEN 10000

static pluginIndicationID_t PECPLUGIN_pluginIndicationIDArray[ARRAY_SIZE];
static int pluginIndicationIDArrayLen = 0;
static int indicationsEnabled = 0;
static int filterActivated = 0;
static pluginData_t *PECPLUGIN_Threads;
static threadData_t *threadData;
static const CMPIBroker *_broker;

static void setEnvPath(char *env, int len);
static void getOptionsAndArguments(int index, char *optionsAndArguments);
static void freeArgumentsArray(char **argv, int len);
static int createArgumentsArray(const char *optionsAndArguments,
                                char **argv,
                                const char *filename,
                                int len);
static void *monitorPECPlugin(void *data);
static int triggerEachPlugin(const char *filterString, const CMPIContext *ctx);
static int triggerAllPlugins(const CMPIContext *ctx);
static int allocnCreatePlugin(const CMPIContext *ctx);
static const char *searchWHEREClause(const char *filterString);
static CMPIStatus PECPluginActivateFilter(CMPIIndicationMI *,
                                             const CMPIContext *,
                                             const CMPISelectExp *,
                                             const char *,
                                             const CMPIObjectPath *,
                                             CMPIBoolean );
static CMPIStatus PECPluginDeActivateFilter(CMPIIndicationMI *,
                                               const CMPIContext *,
                                               const CMPISelectExp *,
                                               const char *,
                                               const CMPIObjectPath *,
                                               CMPIBoolean );
static CMPIStatus PECPluginMustPoll(CMPIIndicationMI *, 
                                       const CMPIContext *, 
                                       const CMPISelectExp *, 
                                       const char *, 
                                       const CMPIObjectPath *);
static CMPIStatus PECPluginAuthorizeFilter(CMPIIndicationMI *, 
                                              const CMPIContext *, 
                                              const CMPISelectExp *, 
                                              const char *, 
                                              const CMPIObjectPath *,
                                              const char *);
static CMPIStatus PECPluginEnableIndications(CMPIIndicationMI *, 
                                          const CMPIContext *);
static CMPIStatus PECPluginDisableIndications(CMPIIndicationMI *, 
                                           const CMPIContext *);
static CMPIStatus PECPluginIndicationCleanup(CMPIIndicationMI *,
                                                const CMPIContext *,
                                                CMPIBoolean );

static CMPIIndicationMIFT indMIFT__=
{
   CMPICurrentVersion,
   CMPICurrentVersion,
   "IndicationPECPlugin",
   PECPluginIndicationCleanup, 
   PECPluginAuthorizeFilter,
   PECPluginMustPoll,
   PECPluginActivateFilter,
   PECPluginDeActivateFilter,
   PECPluginEnableIndications,
   PECPluginDisableIndications,
 };

/*This function sets in  the env string the Environmental Path */
 static void setEnvPath(char *env, int len)
{
    char *ptr;
    char const *pathtxt="PATH=%s";
    int envpathlen;
    
    ptr = getenv("PATH");
    if (ptr && (envpathlen = strlen(ptr) + strlen(pathtxt) + 1) < MAX_STR_LEN)    
    {
        snprintf(env, envpathlen, pathtxt, ptr); 
    }
}

/*Retrieves the options and arguments  loaded in the PECplugin  data array*/
static void getOptionsAndArguments(int index, char *optionsAndArguments)
{
    int opnarglen = strlen(PECPLUGIN_pluginIndicationIDArray[index].optAndArgs) + 1;    
    if (opnarglen < MAX_STR_LEN)
    {
        snprintf(optionsAndArguments, opnarglen, PECPLUGIN_pluginIndicationIDArray[index].optAndArgs);
    }
}

/*Frees the arguments array*/
static void freeArgumentsArray(char **argv, int len)
{
    int i;
    for(i = 0; i < len; i++)
    {
        free(argv[i]);
    }
}

/*Creates an array with the arguments of each plugin-function*/
static int createArgumentsArray(const char *optionsAndArguments,
                         char **argv,
                         const char *filename,
                         int len)
{
    const char *ptr;
    const char *currPos;
    char *nextString;
    int length;
    int i;

    currPos = optionsAndArguments;
    length = strlen(filename) + 1;
    nextString = (char *)malloc(length);
    strncpy(nextString, filename, length);
    nextString[length] = 0;
    argv[0] = nextString;
    for(i = 1; i < len; i++)
    {
        for(;*currPos == ' '; currPos++);
        ptr = strchr(currPos, ' ');
        if (ptr)
        {
            length = ptr - currPos;
            nextString = (char *)malloc(length + 1);
            strncpy(nextString, currPos, length);
            nextString[length] = 0;
            argv[i] = nextString;
            currPos = ptr;
        }
        else
        {
            if (*currPos != 0)
            {
                length = strlen(currPos) + 1;
                nextString = (char *)malloc(length);
                strncpy(nextString, currPos, length);
                nextString[length] = 0;
                argv[i] = nextString;
                return (i + 1);
            }
            else
            {
                return i;
            }
        }
    }
    for(;*currPos == ' '; currPos++);
    return (*currPos != 0)? -1 : len;
}

/*This function runs a thread per Functionality of a plugin loaded in the PECPLUGIN_pluginIndicationIDArray*/
static void *monitorPECPlugin(void *data)
{
    CMPIInstance *instance;
    CMPIStatus status;
    CMPIObjectPath *op;
    char snvalue[MAX_SENTENCE_LEN];
    char *env[MAX_ENV_STR + 1];
    char envPATH[MAX_STR_LEN + 1];
    int pluginProviderPipe[2];
    char buf[MAX_BUF_LEN];
    size_t count = MAX_READ_SIZE;
    int lenread = 0;
    char pluginName[MAX_STR_LEN + 1];
    char pluginPath[MAX_STR_LEN + 1];
    char optionsAndArguments[MAX_STR_LEN + 1];
    char *argv[ARGS_NUM];
    char *ptr;
    int result;
    unsigned char endwait = 0;
    pid_t pidPlugin, pluginResult;
    int respid, pidstatus;
    int deliverInd, index;
    char const *probableCauseStr = NULL;
    char const probableCauseMisConf[] = PROBABLE_CAUSE_PLUG_IN_MISSCONF;
    char const probableCausePlugErr[] = PROBABLE_CAUSE_PLUG_IN_ERROR;
    char const *indicationIdStr = NULL;
    char const indicationIdMisConf[] = INDICATION_ID_PLUG_IN_MISSCONF;
    char const indicationIdPlugErr[] = INDICATION_ID_PLUG_IN_ERROR;


    index = ((threadData_t *)data)->index;
    while(!indicationsEnabled)
    {
        sched_yield();
    }

    ptr = getenv("PEC_PLUGIN_PATH");
    if (ptr == NULL)
    {
        strncpy(pluginPath, PEC_DEFAULT_PLUGIN_PATH, MAX_STR_LEN);
        debugOutput("Missing PEC_PLUGIN_PATH env var, setting to default in: %s\n", pluginPath);
    }
    else
    {
        strncpy(pluginPath, ptr, MAX_STR_LEN);
    }

    strncpy(pluginName, pluginPath, MAX_STR_LEN);
    strncat(pluginName, "/", MAX_STR_LEN);
    strncat(pluginName, PECPLUGIN_pluginIndicationIDArray[index].pluginName, MAX_STR_LEN);
    CBAttachThread(_broker, PECPLUGIN_Threads[index].newContext);
    PECPLUGIN_Threads[index].threadRes = 0;
    strncpy(snvalue, "CIM_AlertIndication", MAX_SENTENCE_LEN);
    getOptionsAndArguments(index, optionsAndArguments);
    result = createArgumentsArray(optionsAndArguments, argv, pluginName, ARGS_NUM - 1);
    if (result == -1)
    {
        freeArgumentsArray(argv, ARGS_NUM -1);
        debugOutput("Error: too many arguments\n");
        PECPLUGIN_Threads[index].threadRes = -1;
        return (void *) &PECPLUGIN_Threads[index].threadRes;
    }
    else
    {
        int j;
        argv[result] = NULL;
        for(j=0; argv[j] != NULL; j++) debugOutput("Res: %d - ARGUM: %s\n", result, argv[j]);
    }
    env[0] = envPATH;
    env[1] = NULL;
    setEnvPath(env[0], MAX_STR_LEN + 1);
    while(!PECPLUGIN_Threads[index].endThread)
    {
        deliverInd = 0;
        sleep(PECPLUGIN_pluginIndicationIDArray[index].frequency);
        if (!PECPLUGIN_Threads[index].endThread)
        {
            if (pipe(pluginProviderPipe) == -1)
            {
                debugOutput("Pipe creation failed: %d\n", errno);
            }        
            endwait = 0;
            pidPlugin = fork();
            if (pidPlugin == 0)
            {
                int fdout;
            
                fdout = dup(1);
                if (fdout == -1)
                {
                    debugOutput("Error dup stdout: %d\n", errno);
                }
                close(1);
                if (dup(pluginProviderPipe[1]) == -1)
                {
                   debugOutput("Error dup pluginProviderPipe: %d\n", errno);
                }
                close(pluginProviderPipe[1]);
                close(pluginProviderPipe[0]);
    
                if (execve(pluginName, argv, env) == -1)
                {
                    close(1);
                    dup(fdout);
                    debugOutput("ERROR calling execve: %d\n", errno);
                    close(fdout);
                    exit(255);

                }
                close(fdout);
                exit(0);
            }
            else if (pidPlugin == -1)
            {
                debugOutput("ERROR calling fork: %d - thread %d\n", errno, index);
            }
            else
            {
                while(!endwait)
                {
                    respid = waitpid(pidPlugin, &pidstatus, 0);
                    if (respid == pidPlugin)
                    {
                        if (WIFEXITED(pidstatus)) 
                        {
                            pluginResult = WEXITSTATUS(pidstatus);
                            endwait = 1;
                        }
                        else
                        {
                            if (!WIFSTOPPED(pidstatus) && !WIFCONTINUED(pidstatus))
                            {
                                endwait = 1;
                                pluginResult = 255;
                            }
                        }
                    }
                    else
                    {
                        debugOutput("waitpid returned != respid: %d\n", respid);
                        if (errno != EINTR)
                        {
                            endwait = 1;
                            pluginResult = 255;
                        }
                    }
                }
                debugOutput("Thread index: %d - Forked process ID %d which terminated: %d\n", index, pidPlugin, pluginResult);
                close(pluginProviderPipe[1]);
                if (pluginResult != 255)
                {
                    lenread = read(pluginProviderPipe[0], (void *)buf, count);
                    if (lenread == -1)
                    {
                        debugOutput("Error reading: %d\n", errno);
                    }
                    else
                    {
                        buf[lenread] = 0;
                        debugOutput("READ %d BYTES: %s\n", lenread, buf);
                        debugOutput("Plugin Result: %d\n", pluginResult);
                        if (pluginResult != 0)
                        {
                            deliverInd = 1;
                        }
                    }
                }
                close(pluginProviderPipe[0]);
                if ((deliverInd == 1 || pluginResult == 255 ) && indicationsEnabled && !PECPLUGIN_Threads[index].endThread)
                {
                    CMPIDateTime *dtptr;
                    CMPIUint16 perceivedSeverity = 0;
                    debugOutput("Indication to deliver. Plugin result: %d\n", pluginResult);

                    switch(pluginResult)
                    {
                        case 1:
                            perceivedSeverity = 3;
                            probableCauseStr = PECPLUGIN_pluginIndicationIDArray[index].probableCause;
                            indicationIdStr = PECPLUGIN_pluginIndicationIDArray[index].indicationIdentifier;
                        break;
                        case 2: 
                            perceivedSeverity = 6;
                            probableCauseStr = PECPLUGIN_pluginIndicationIDArray[index].probableCause;
                            indicationIdStr = PECPLUGIN_pluginIndicationIDArray[index].indicationIdentifier;
                        break;
                        case 255:
                            perceivedSeverity = 3;
                            probableCauseStr = probableCausePlugErr;
                            indicationIdStr = indicationIdPlugErr;
                        break;
                        default: 
                            perceivedSeverity = 2;
                            probableCauseStr = probableCauseMisConf;
                            indicationIdStr = indicationIdMisConf;
                        break;
                    }

                    op = CMNewObjectPath(_broker, "root/interop", "CIM_AlertIndication", &status);
                    instance = CMNewInstance(_broker, op, &status);
                    CMSetProperty(instance, "ClassName", snvalue, CMPI_chars);
                    dtptr = _broker->eft->newDateTime(_broker, &status);
                    if ((status.rc != CMPI_RC_OK) || (dtptr == NULL))
                    {
                        debugOutput("Could not get DateTime\n");
                    }
                    else
                    {
                        CMSetProperty(instance, "IndicationTime", (CMPIValue*)&dtptr, CMPI_dateTime);
                    }
                    CMSetProperty(instance, "PerceivedSeverity", &perceivedSeverity, CMPI_uint16);
                    CMSetProperty(instance, "IndicationIdentifier", indicationIdStr, CMPI_chars);
                    CMSetProperty(instance, "ProbableCauseDescription", probableCauseStr, CMPI_chars);
                    if (indicationsEnabled && !PECPLUGIN_Threads[index].endThread)
                    {
                        status = CBDeliverIndication(_broker, PECPLUGIN_Threads[index].newContext, "root/interop", instance);
                        if (status.rc != CMPI_RC_OK)
                        {
                            debugOutput("Error delivering indication\n");
                        }
                    }
                }
            }
        }
        sched_yield();
    }
    debugOutput("Thread terminated: %d\n", index);
    PECPLUGIN_Threads[index].threadStarted = 0;
    return (void *) &PECPLUGIN_Threads[index].threadRes;
}

/*This function allocate memory for each plugin an run it*/
static int allocnCreatePlugin(const CMPIContext *ctx)
{
    int i;
    int res;

    PECPLUGIN_Threads = (pluginData_t *)malloc(pluginIndicationIDArrayLen * sizeof(pluginData_t));
    if (PECPLUGIN_Threads == NULL)
    {
        debugOutput("Could not allocate memory for PECPLUGIN_Threads\n");
        return -1;
    }
    bzero(PECPLUGIN_Threads, pluginIndicationIDArrayLen * sizeof(pluginData_t));
    threadData = (threadData_t *)malloc(pluginIndicationIDArrayLen * sizeof(threadData_t));
    if (threadData == NULL)
    {
        debugOutput("Could not allocate memory for PECPLUGIN_Threads\n");
        return -1;
    }
    for(i = 0; i < pluginIndicationIDArrayLen; i++)
    {
        if (PECPLUGIN_Threads[i].threadStarted == 0)
        {
            bzero(&PECPLUGIN_Threads[i], sizeof(pluginData_t));
            pthread_attr_init(&PECPLUGIN_Threads[i].monitoringThreadAttr);
            pthread_attr_setdetachstate(&PECPLUGIN_Threads[i].monitoringThreadAttr, PTHREAD_CREATE_DETACHED);
            PECPLUGIN_Threads[i].newContext = CBPrepareAttachThread(_broker, ctx);
            PECPLUGIN_Threads[i].endThread = 0;
            threadData[i].index = i; 
            res = pthread_create(&PECPLUGIN_Threads[i].monitoringThread, 
                                 &PECPLUGIN_Threads[i].monitoringThreadAttr, 
                                 monitorPECPlugin, &threadData[i]);
            if (res != 0)
            {
                debugOutput("Error creating Thread: %d\n", res);
            }
            PECPLUGIN_Threads[i].threadStarted = 1;
        }
    }
    return 0;
}

/*This function enumerates the selected plug-ins and runs it*/
static int triggerEachPlugin(const char *filterString, const CMPIContext *ctx)
{
    int res;

    res = getFilteredPluginsData(_broker, ctx, PECPLUGIN_pluginIndicationIDArray, filterString, &pluginIndicationIDArrayLen);
    res = allocnCreatePlugin(ctx);
    return res;
}

/*This function enumerates all the plug-ins and runs it*/
static int triggerAllPlugins(const CMPIContext *ctx)
{
    int res;

    res = getAllPluginsData(_broker, ctx, PECPLUGIN_pluginIndicationIDArray, ARRAY_SIZE, &pluginIndicationIDArrayLen);
    res = allocnCreatePlugin(ctx);
    return res;
}


static const char *searchWHEREClause(const char *filterString)
{
    return strstr(filterString, WHERE_KEY);
}

/*This is the filter for the PEC PLUGIN*/
static CMPIStatus PECPluginActivateFilter(CMPIIndicationMI *thisMI,
                                             const CMPIContext *ctx,
                                             const CMPISelectExp *filter,
                                             const char *eventType,
                                             const CMPIObjectPath *classPath,
                                             CMPIBoolean firstActivation)                                              
{
    CMPIString *classpath;
    CMPIStatus status;
    CMPIString *filterString;
    char filterChar[MAX_FILTER_LEN + 1];
    const char *whereClause;
    
    if (filterActivated == 1)
    {
        debugOutput("Filter already Active\n");
        status.rc = CMPI_RC_ERR_FAILED;
        status.msg = NULL;
        return status;
    }

    filterString = filter->ft->getString(filter, &status);
    strncpy(filterChar, filterString->ft->getCharPtr(filterString, &status), MAX_FILTER_LEN);
    filterChar[MAX_FILTER_LEN] = 0;
    if (strncmp(eventType, ALERT_CLASS_NAME, MAX_SENTENCE_LEN) == 0)
    {
        classpath = classPath->ft->getClassName(classPath, &status);
        whereClause = searchWHEREClause(filterChar);
        if (whereClause == NULL)
        {
            triggerAllPlugins(ctx);
        }
        else
        {
            triggerEachPlugin(whereClause, ctx);
        }
    }
    filterActivated = 1; 
    debugOutput("Filter Activated\n");
    status.rc = CMPI_RC_OK;
    status.msg = NULL;
    return status;
}

/*Thius function deactivates the filters*/
static CMPIStatus PECPluginDeActivateFilter(CMPIIndicationMI *thisMI,
                                               const CMPIContext *ctx,
                                               const CMPISelectExp *filter,
                                               const char *eventType,
                                               const CMPIObjectPath *classPath,
                                               CMPIBoolean lastActivation)                                              
{
    CMPIString *classpath;
    CMPIStatus status;
    int i;
    classpath = classPath->ft->getClassName(classPath, &status);
    debugOutput("Deactivating %d Filters\n", pluginIndicationIDArrayLen);
    if (lastActivation == 1)
    {
        for(i = 0; i < pluginIndicationIDArrayLen; i++)
        {
           PECPLUGIN_Threads[i].endThread = 1;
        }

        for(i = 0; i < pluginIndicationIDArrayLen; i++)
        {
            while(PECPLUGIN_Threads[i].threadStarted == 1) 
            {
                sched_yield();
            }
            debugOutput("Filter %d deactivated\n", i+1);
        }
    }
    debugOutput("Ended deactivating Filters\n");
    filterActivated = 0;
    status.rc = CMPI_RC_OK;
    status.msg = NULL;
    return status;
}


static CMPIStatus PECPluginMustPoll(CMPIIndicationMI *mi, 
                                       const CMPIContext *ctx, 
                                       const CMPISelectExp *filter, 
                                       const char *indType, 
                                       const CMPIObjectPath *classPath) 
{
    CMPIStatus rc = {CMPI_RC_ERR_FAILED, NULL};
    return rc;
}


static CMPIStatus PECPluginAuthorizeFilter(CMPIIndicationMI *mi, 
                                        const CMPIContext *ctx, 
                                        const CMPISelectExp *filter, 
                                        const char *eventType, 
                                        const CMPIObjectPath *classPath,
                                        const char *owner) 
{
    CMReturn(CMPI_RC_OK);
}


static CMPIStatus PECPluginEnableIndications(CMPIIndicationMI *mi, 
                                          const CMPIContext *ctx) 
{
    indicationsEnabled  = 1;
    
}

static CMPIStatus PECPluginDisableIndications(CMPIIndicationMI *mi, 
                                           const CMPIContext *ctx)
{
    indicationsEnabled  = 0;
}

static CMPIStatus PECPluginIndicationCleanup(CMPIIndicationMI *mi,
                                                const CMPIContext *ctx,
                                                CMPIBoolean terminating)
{
    debugOutput("PECPluginCleanup\n");
    CMReturn(CMPI_RC_OK);
}

CMPI_EXTERN_C CMPIIndicationMI* PECPlugin_Create_IndicationMI(const CMPIBroker *brkr,const CMPIContext *ctx, CMPIStatus *rc)
{ 
    static CMPIIndicationMI mi={NULL, &indMIFT__, }; 
    _broker=brkr; 
    CMNoHook; 
    return &mi; 
}
