/*******************************************************************************
 * Copyright (c) 2008 Intel Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Rodolfo Kohn - initial API and implementation
 *    Marcelo Lorenzatti, Miriam Estela - Management of multiple configurations
      and filters .
 *******************************************************************************
 Product name: Plugin Extension for CIM  
 Creation date: 1-Nov-2007
 Description: 
 This file is the responsible of enumerating and showing all the plugins and its
 related functionalities

 Modification History
 02/20/2008      Marcelo Lorenzati, Miriam Estela    Fixed to read more than one
                 functionality per plugin
********************************************************************************/

#define CMPI_PLATFORM_LINUX_GENERIC_GNU

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include "cmpidt.h"
#include "cmpift.h"
#include "cmpimacs.h"    
#include "pec_plugins.h"

#define NAMESPACE                    "root/cimv2"
#define PLUGIN_CLASS                 "PEC_Plugin"
#define PLUGIN_FUNCTIONALITY_CLASS   "PEC_PluginFunctionalitySettingData"
#define PLUGIN_FUNC_INDICATION_CLASS "PEC_RelatedIndication"
#define RELATED_PROPERTY_CLASS       "PEC_RelatedProperty"
#define PLUGIN_FUNC_ASSOCIATION      "CIM_ElementSettingData"
#define FUNC_IND_ASSOCIATION         "PEC_FunctionalityIndicationSupport"
#define INSTANCE_ID_LEN STRING_LEN
#define SELECT_FUNCTIONALITY "SELECT * FROM %s WHERE InstanceID='%s'"
#define QUERY_PLUGIN_FUNC_LEN (sizeof(SELECT_FUNCTIONALITY) + sizeof(PLUGIN_FUNCTIONALITY_CLASS) + INSTANCE_ID_LEN)
#define SELECT_INDICATION "SELECT * FROM %s WHERE IndicationIdentifier='%s'"
#define SELECT_INDICATION_FILTER "SELECT * FROM %s %s"
#define QUERY_FUNC_INDIC_LEN (sizeof(SELECT_INDICATION) + sizeof(PLUGIN_FUNC_INDICATION_CLASS) + INSTANCE_ID_LEN)

static void getIndData(CMPIInstance *instance, 
                       pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                       int arrayInd);

static void getIndication(const CMPIBroker *broker, 
                          const CMPIContext *ctx,
                          CMPIObjectPath *objectpath1,
                          pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                          int arrayInd);

static void getFuncIndication(const CMPIBroker *broker, 
                              const CMPIContext *ctx,
                              CMPIInstance *instance, 
                              pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                              int arrayInd);

static void getFuncData(CMPIInstance *instance, 
                        pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                        int arrayInd,
                        functionalityMode_t *mode);

static void getFunc(const CMPIBroker *broker, 
                    const CMPIContext *ctx,
                    CMPIObjectPath *objectpath1,
                    pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                    int arrayInd);

static int compareObjectPathString(CMPIObjectPath *objectpath1, CMPIObjectPath *objectpath2, const char *stringCompare);

static void getInstanceFunc(const CMPIBroker *broker, 
                            const CMPIContext *ctx,
                            CMPIInstance *instance, 
                            pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                            int *arrayInd);
static int getOpAssocElemSetData(const CMPIBroker *broker,
                                 const CMPIContext *ctx, 
                                 const char *nameAssoc,
                                 const char *nameAtribCompare, 
                                 const char *nameAtrib1,
                                 const char *nameAtrib2,
                                 CMPIObjectPath *opFunctionality, 
                                 pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                                 int localInd);
static int getOpFromAssocIndSup(const CMPIBroker *broker,
                                const CMPIContext *ctx, 
                                const char *nameAssoc,
                                const char *nameAtribCompare, 
                                const char *nameAtrib1,
                                const char *nameAtrib2,
                                CMPIObjectPath *opIndication, 
                                pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                                int localInd);
static void enumerateAllData(const CMPIBroker *broker, 
                             const CMPIContext *ctx,
                             pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                             int arraySize, 
                             int *pluginIndicationIDArrayLen);
static void enumerateFilteredData(const CMPIBroker *broker, 
                                  const CMPIContext *ctx,
                                  pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                                  const char *filterString, 
                                  int *pluginIndicationIDArrayLen);
static void showElements(pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                         int pluginIndicationIDArrayLen);
int getAllPluginsData(const CMPIBroker *broker,
                      const CMPIContext *ctx,
                      pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                      int arraySize, 
                      int *pluginIndicationIDArrayLen);
int aetFilteredPluginsData(const CMPIBroker *broker,
                             const CMPIContext *ctx,
                             pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                             const char *filterString,
                             int *pluginIndicationIDArrayLen);


#ifdef DEBUG 
static int debugLevel = 1;
#else
static int debugLevel = 0;
#endif

int debugOutput(char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    if (debugLevel)
    {
        vprintf(fmt, ap);
    }
    va_end(ap);

    return 0;
}

/*This function shows the elements and configuration loaded in the PECPLUGIN_pluginIndicationIDArray*/            
static void showElements(pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                         int pluginIndicationIDArrayLen)
{
    int i;
    for(i = 0; i < pluginIndicationIDArrayLen; i++)
    {
        debugOutput("\nNew Element\n");
        debugOutput("Plugin Name: %s\n", PECPLUGIN_pluginIndicationIDArray[i].pluginName);
        debugOutput("Indication Identifier: %s\n", PECPLUGIN_pluginIndicationIDArray[i].indicationIdentifier);
        debugOutput("Frequency: %d\n", PECPLUGIN_pluginIndicationIDArray[i].frequency);
        debugOutput("Options & Arguments: %s\n", PECPLUGIN_pluginIndicationIDArray[i].optAndArgs);
    }
    debugOutput("\n\n");
}


/*This function enumerates and shows all the data of the selected plugins*/
int getFilteredPluginsData(const CMPIBroker *broker,
                             const CMPIContext *ctx,
                             pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                             const char *filterString,
                             int *pluginIndicationIDArrayLen)
{
    enumerateFilteredData(broker, ctx, PECPLUGIN_pluginIndicationIDArray, filterString, pluginIndicationIDArrayLen);
    showElements(PECPLUGIN_pluginIndicationIDArray, *pluginIndicationIDArrayLen);
    return 0;
}

/*This function enumerates and shows all the data of all the plugins */
int getAllPluginsData(const CMPIBroker *broker,
                      const CMPIContext *ctx,
                      pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                      int arraySize, 
                      int *pluginIndicationIDArrayLen)
{
    enumerateAllData(broker, ctx, PECPLUGIN_pluginIndicationIDArray, arraySize, pluginIndicationIDArrayLen);
    showElements(PECPLUGIN_pluginIndicationIDArray, *pluginIndicationIDArrayLen);
    return 0;
}

/*This Function compares a single Key string of two object path*/
static int compareObjectPathString(CMPIObjectPath *objectpath1, CMPIObjectPath *objectpath2, const char *stringCompare)
{
    CMPIStatus status;
    const char *ch1, *ch2;

    CMPIData data1 = objectpath1->ft->getKey(objectpath1, stringCompare, &status);
    if (status.msg) CMRelease(status.msg);
    CMPIData data2 = objectpath2->ft->getKey(objectpath2, stringCompare, &status);
    if (status.msg) CMRelease(status.msg);
    ch1 = data1.value.string->ft->getCharPtr(data1.value.string, &status);
    if (status.msg) CMRelease(status.msg);
    ch2 = data2.value.string->ft->getCharPtr(data2.value.string, &status);
    if (status.msg) CMRelease(status.msg);

    return (strncmp(ch1, ch2, STRING_LEN) == 0);
}

/*This Function gets the object path of an association of a given object path of a Functionality*/
static int getOpAssocElemSetData(const CMPIBroker *broker,
                                 const CMPIContext *ctx, 
                                 const char *nameAssoc,
                                 const char *nameAtribCompare, 
                                 const char *nameAtrib1,
                                 const char *nameAtrib2,
                                 CMPIObjectPath *opFunctionality, 
                                 pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                                 int localInd)
{
    CMPIObjectPath *opElemSetData = broker->eft->newObjectPath(broker, NAMESPACE, nameAssoc, NULL);
    CMPIEnumeration *enumElemSetData;
    CMPIStatus status;
    int result = 1;
   
    enumElemSetData = broker->bft->enumerateInstances(broker, ctx, opElemSetData, NULL, &status);
    if (!status.rc)
    {
        CMPIData dataPropSettingData, dataPropManagedElement;
        CMPIStatus status3;
        CMPIData dataElemSetData;
        int assocFind=0;

        if (status.msg) CMRelease(status.msg);
        while (enumElemSetData->ft->hasNext(enumElemSetData, NULL) && (!assocFind))
        {
            dataElemSetData = enumElemSetData->ft->getNext(enumElemSetData, NULL);
            if (status.msg) CMRelease(status.msg);
            dataPropSettingData = dataElemSetData.value.inst->ft->getProperty(dataElemSetData.value.inst, nameAtrib1, &status);
            if (!status.rc)
            {   
                if (status.msg) CMRelease(status.msg);
                CMPIObjectPath *opSettingData = dataPropSettingData.value.ref;
                if(compareObjectPathString(opSettingData, opFunctionality, nameAtribCompare))
                {
                    assocFind = 1;
                    dataPropManagedElement = dataElemSetData.value.inst->ft->getProperty(dataElemSetData.value.inst, nameAtrib2, &status);
                    if (!status.rc)
                    {  
                        CMPIInstance *instancePlugin;
                        CMPIObjectPath *objectpathPlugin = dataPropManagedElement.value.ref; 
                        if (status.msg) CMRelease(status.msg);
                        CMPIData dataPlugin = opElemSetData->ft->getKey(objectpathPlugin, "Name", &status);
                        if (!status.rc) 
                        {
                            result = 0;
                            if (status.msg) CMRelease(status.msg);
                            const char *namePlugin = dataPlugin.value.string->ft->getCharPtr(dataPlugin.value.string, &status);
                            int longVar = strlen(namePlugin);
                            if (longVar > STRING_LEN) longVar = STRING_LEN;
                            strncpy(PECPLUGIN_pluginIndicationIDArray[localInd].pluginName, namePlugin, longVar);
                        }
                        else
                        {
                            debugOutput("Error in getProperty, rc=%d, msg=%s\n", 
                                    status.rc, (status.msg)? (char *)status.msg->hdl : NULL);  
                            result = 1;
                        }
                        if (objectpathPlugin) CMRelease(objectpathPlugin);
                    }
                    else
                    {          
                        debugOutput("Error in getProperty, rc=%d, msg=%s\n", 
                                status.rc, (status.msg)? (char *)status.msg->hdl : NULL);  
                        result = 1;
                    }
                }
                if (opSettingData)  CMRelease(opSettingData);
            }
            else
            {          
                debugOutput("Error in getProperty, rc=%d, msg=%s\n", 
                        status.rc, (status.msg)? (char *)status.msg->hdl : NULL);  
                result = 1;
                assocFind = 1;
            }
        }
    }
    else
    {          
        debugOutput("Error in getOpAssocElemSetData, rc=%d, msg=%s\n", 
                status.rc, (status.msg)? (char *)status.msg->hdl : NULL);      
    }

    if (status.msg) CMRelease(status.msg); 
    if (enumElemSetData) CMRelease(enumElemSetData);
    if (opElemSetData)  CMRelease(opElemSetData);

    return result;                    
 }

/*This Function gets the association from an Indication in PEC_Relatedindication*/
static int getOpFromAssocIndSup(const CMPIBroker *broker,
                                const CMPIContext *ctx, 
                                const char *nameAssoc,
                                const char *nameAtribCompare, 
                                const char *nameAtrib1,
                                const char *nameAtrib2,
                                CMPIObjectPath *opIndication, 
                                pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                                int localInd)
{
    CMPIData dataEnumAssoc, dataPropIndicator, dataPropFunctionality;
    CMPIEnumeration *enumAssoc;
    CMPIStatus status;
    CMPIObjectPath *opAssociation = broker->eft->newObjectPath(broker, NAMESPACE, nameAssoc, NULL);
    int result = 1;
    int assocFind = 0;

    enumAssoc = broker->bft->enumerateInstances(broker, ctx, opAssociation, NULL, &status);
    if (!status.rc)
    {
        if (status.msg) CMRelease(status.msg);
        while (enumAssoc->ft->hasNext(enumAssoc, NULL) && !assocFind)
        {
            dataEnumAssoc = enumAssoc->ft->getNext(enumAssoc, NULL);
            dataPropIndicator = dataEnumAssoc.value.inst->ft->getProperty(dataEnumAssoc.value.inst, nameAtrib1, &status);
            if (!status.rc)
            {          
                if (status.msg) CMRelease(status.msg);
                CMPIObjectPath *opAssocInd = dataPropIndicator.value.ref;
                if(compareObjectPathString(opAssocInd, opIndication, nameAtribCompare))
                {
                    assocFind = 1;
                    dataPropFunctionality = dataEnumAssoc.value.inst->ft->getProperty(dataEnumAssoc.value.inst, nameAtrib2, &status);
                    if (!status.rc)
                    {          
                        if (status.msg) CMRelease(status.msg);
                        CMPIInstance *instanceFunctionality;
                        CMPIObjectPath *opFunctionality = dataPropFunctionality.value.ref; 
                        instanceFunctionality = broker->bft->getInstance(broker, ctx, opFunctionality, NULL, &status);
                        if (!status.rc) 
                        {
                            CMPIInstance *instancePlugin;
                            CMPIObjectPath *opSelectedplugin;
                            CMPIData dataPropArg = instanceFunctionality->ft->getProperty(instanceFunctionality, "OptionsAndArguments", &status);
                            const char * dataArg = dataPropArg.value.string->ft->getCharPtr(dataPropArg.value.string, &status);
                            int longVar = strlen(dataArg);

                            if (longVar > STRING_LEN) longVar = STRING_LEN;
                            strncpy(PECPLUGIN_pluginIndicationIDArray[localInd].optAndArgs, dataArg, longVar);
                            opSelectedplugin = instanceFunctionality->ft->getObjectPath(instanceFunctionality, NULL);
                            result = getOpAssocElemSetData(broker, ctx, PLUGIN_FUNC_ASSOCIATION, "InstanceID", "SettingData", 
                                                          "ManagedElement", opSelectedplugin, PECPLUGIN_pluginIndicationIDArray, localInd);
                        }
                        else
                        {
                            debugOutput("Error in getInstance, rc=%d, msg=%s\n", 
                                    status.rc, (status.msg)? (char *)status.msg->hdl : NULL);  
                            result = 1;
                        }
                    }
                    else
                    {
                        debugOutput("Error in getProperty, rc=%d, msg=%s\n", 
                                status.rc, (status.msg)? (char *)status.msg->hdl : NULL);  
                        result = 1;
                    }
                }
            }
            else
            {
                debugOutput("Error in getProperty, rc=%d, msg=%s\n", 
                        status.rc, (status.msg)? (char *)status.msg->hdl : NULL);  
                assocFind = 1;
                result = 1;               
            }   
         }
    }
    else
    {          
        result = 1;
        debugOutput("Error in getOpFromAssocIndSup, rc=%d, msg=%s\n", 
               status.rc, (status.msg)? (char *)status.msg->hdl : NULL);
    }

 if (status.msg) CMRelease(status.msg); 
 if (enumAssoc) CMRelease(enumAssoc); 
 return result;                    
 }

/*This function conects to the broker and enumerate all the instances related to PEC_PLUGINS*/
static void enumerateAllData(const CMPIBroker *broker, 
                             const CMPIContext *ctx,
                             pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                             int arraySize, 
                             int *pluginIndicationIDArrayLen)
{
    CMPIObjectPath *opPECPlugin;
    CMPIStatus status;
    CMPIEnumeration *enumPECPlugin;
    int localInd = 0;
    
    *pluginIndicationIDArrayLen = 0;
    opPECPlugin = broker->eft->newObjectPath(broker, NAMESPACE, PLUGIN_CLASS, NULL);
    enumPECPlugin = broker->bft->enumerateInstances(broker, ctx, opPECPlugin, NULL, &status);
    if (!status.rc)
    {
        while (enumPECPlugin->ft->hasNext(enumPECPlugin, NULL) && (localInd < arraySize))
        {
            CMPIData data = enumPECPlugin->ft->getNext(enumPECPlugin, NULL);
            CMPIData propertyName = data.value.inst->ft->getProperty(data.value.inst, "Name", &status);
            int longVar = strlen((propertyName.value.string)->ft->getCharPtr(propertyName.value.string, &status));
            if (longVar > STRING_LEN)
                longVar = STRING_LEN;
            strncpy(PECPLUGIN_pluginIndicationIDArray[localInd].pluginName, 
                   (propertyName.value.string)->ft->getCharPtr(propertyName.value.string, &status), longVar);
            getInstanceFunc(broker, ctx, data.value.inst, PECPLUGIN_pluginIndicationIDArray, &localInd);
        }
        debugOutput("Created %d elements\n", localInd);
        *pluginIndicationIDArrayLen = localInd;
    }
    else
    {
        debugOutput("Error in enumerate instances in PEC_Plugin, rc=%d, msg=%s\n", 
               status.rc, (status.msg)? (char *)status.msg->hdl : NULL);  
    }
    if (enumPECPlugin) CMRelease(enumPECPlugin);
    if (opPECPlugin) CMRelease(opPECPlugin);
    if (status.msg) CMRelease(status.msg);
}

static void enumerateFilteredData(const CMPIBroker *broker, 
                                  const CMPIContext *ctx,
                                  pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                                  const char *filterString, 
                                  int *pluginIndicationIDArrayLen)
{
    CMPIStatus status;
    CMPIEnumeration *enumIndication;
    CMPIObjectPath *opPECIndication;
    CMPIObjectPath *opSelectedIndication;
    CMPIObjectPath *objectpathFunc;
    char query[QUERY_FUNC_INDIC_LEN + 1];
    int longVar;

    *pluginIndicationIDArrayLen = 0; 
    opPECIndication = broker->eft->newObjectPath(broker, NAMESPACE, PLUGIN_FUNC_INDICATION_CLASS, NULL);
    longVar = strlen(filterString);
    if (longVar > INSTANCE_ID_LEN) debugOutput("Maximum query size lenght exceeded\n");
    snprintf(query, QUERY_FUNC_INDIC_LEN, SELECT_INDICATION_FILTER, PLUGIN_FUNC_INDICATION_CLASS, filterString);
    enumIndication = broker->bft->execQuery(broker, ctx, opPECIndication, query, "WQL", &status);
    if (!status.rc)
    {   
        while (enumIndication->ft->hasNext(enumIndication, NULL)) 
        {
            CMPIData dataIndication = enumIndication->ft->getNext(enumIndication, NULL);
            opSelectedIndication = dataIndication.value.inst->ft->getObjectPath(dataIndication.value.inst, NULL);
            getIndData(dataIndication.value.inst, PECPLUGIN_pluginIndicationIDArray, *pluginIndicationIDArrayLen);
            int res = getOpFromAssocIndSup(broker, ctx, FUNC_IND_ASSOCIATION, "IndicationIdentifier", "Indication", 
                                   "Functionality", opSelectedIndication, PECPLUGIN_pluginIndicationIDArray, *pluginIndicationIDArrayLen);
            (*pluginIndicationIDArrayLen)++;
        }
    }
    else  debugOutput("Error on execQuery from enumerateFilteredData, rc=%d, msg=%s\n", 
                  status.rc, (status.msg)? (char *)status.msg->hdl : NULL);

 
    if (enumIndication) CMRelease(enumIndication);
    if (status.msg) CMRelease(status.msg);

}

/*This function checks plugin and calls getFunc to get the functionality of an Instance*/
static void getInstanceFunc(const CMPIBroker *broker, 
                            const CMPIContext *ctx,
                            CMPIInstance *instance, 
                            pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                            int *arrayInd)
{
    CMPIStatus status;
    CMPIEnumeration *enumPluginFuncAssoc;
    CMPIObjectPath *opInput = instance->ft->getObjectPath(instance, NULL);
    CMPIObjectPath *opPluginFuncAssoc;
    CMPIObjectPath *opManagedElement;
    CMPIData dataEnumPluginFuncAssoc;
    CMPIData propertyManagedElem;
    CMPIData propertySettingData; 
    int localInd = 0;

    opPluginFuncAssoc = broker->eft->newObjectPath(broker, NAMESPACE, PLUGIN_FUNC_ASSOCIATION, NULL);
    enumPluginFuncAssoc = broker->bft->enumerateInstances(broker, ctx, opPluginFuncAssoc, NULL, &status);
    if (!status.rc) 
    { 
        while (enumPluginFuncAssoc->ft->hasNext(enumPluginFuncAssoc, NULL)) 
        {
            dataEnumPluginFuncAssoc = enumPluginFuncAssoc->ft->getNext(enumPluginFuncAssoc, NULL);
            propertyManagedElem = dataEnumPluginFuncAssoc.value.inst->ft->getProperty(dataEnumPluginFuncAssoc.value.inst, "ManagedElement", &status);
            opManagedElement = propertyManagedElem.value.ref;
            if (compareObjectPathString(opInput, opManagedElement, "Name"))
            {   
                propertySettingData = dataEnumPluginFuncAssoc.value.inst->ft->getProperty(dataEnumPluginFuncAssoc.value.inst, "SettingData", &status);
                opManagedElement = propertySettingData.value.ref;
                if (localInd > 0)
                {
                    strcpy(PECPLUGIN_pluginIndicationIDArray[*arrayInd].pluginName, 
                            PECPLUGIN_pluginIndicationIDArray[(*arrayInd)-1].pluginName);
                }
                getFunc(broker, ctx, opManagedElement, PECPLUGIN_pluginIndicationIDArray, *arrayInd);
                (*arrayInd) = (*arrayInd) + 1;
                localInd++;
            }
        }

    }
    else
    {
        debugOutput("Error in enumerate instances in CIM_ElementSettingData, rc=%d, msg=%s\n", 
               status.rc, (status.msg)? (char *)status.msg->hdl : NULL);      
    }
    if (enumPluginFuncAssoc) CMRelease(enumPluginFuncAssoc);
    if (opPluginFuncAssoc) CMRelease(opPluginFuncAssoc);
    if (status.msg) CMRelease(status.msg);
}

/*This function gets the functionalities of an Instance and calls getFuncIndication to gets it indications*/
static void getFunc(const CMPIBroker *broker, 
                    const CMPIContext *ctx,
                    CMPIObjectPath *objectpath1,
                    pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                    int arrayInd)
{
    CMPIStatus status;
    CMPIEnumeration *enumPluginFunctionality;
    CMPIData keyInstanceId;
    functionalityMode_t mode;
    char query[QUERY_PLUGIN_FUNC_LEN + 1];
    int longVar;

    keyInstanceId = objectpath1->ft->getKey(objectpath1, "InstanceID", &status);
    longVar = strlen((char *)((keyInstanceId.value.string)->hdl));
    if (longVar > INSTANCE_ID_LEN) debugOutput("Error in Instance Identif.");
    snprintf(query, QUERY_PLUGIN_FUNC_LEN, SELECT_FUNCTIONALITY, 
             PLUGIN_FUNCTIONALITY_CLASS, (char *)((keyInstanceId.value.string)->hdl));
    enumPluginFunctionality = broker->bft->execQuery(broker, ctx, objectpath1, query, "WQL", &status);
    if (!status.rc)
    {
        while (enumPluginFunctionality->ft->hasNext(enumPluginFunctionality, NULL)) 
        {
            CMPIData dataPluginFunctionality = enumPluginFunctionality->ft->getNext(enumPluginFunctionality, NULL);
            getFuncData(dataPluginFunctionality.value.inst, PECPLUGIN_pluginIndicationIDArray, arrayInd, &mode);
            if (mode == INDICATION || mode == ALL)
            {
                getFuncIndication(broker, ctx, dataPluginFunctionality.value.inst, PECPLUGIN_pluginIndicationIDArray, arrayInd);
            }     
        }
    }
    else
    {
        debugOutput("Error in execQuery in PEC_PluginFunctionalitySettingData, rc=%d, msg=%s\n", 
               status.rc, (status.msg)? (char *)status.msg->hdl : NULL);      
    }
    if (enumPluginFunctionality) CMRelease(enumPluginFunctionality);
    if (status.msg) CMRelease(status.msg);
}

/*This function gets the Data of a functionality*/
static void getFuncData(CMPIInstance *instance, 
                        pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                        int arrayInd,
                        functionalityMode_t *mode)
{
    CMPIStatus status;
    CMPIData propertyOptArguments;
    CMPIData propertyMode;
    CMPIString *valstring;

    CMPIObjectPath *objectpath1 = instance->ft->getObjectPath(instance, NULL);
    propertyOptArguments = instance->ft->getProperty(instance, "OptionsAndArguments", &status);
    valstring = propertyOptArguments.value.string;
    int longVar = strlen((char *)valstring->hdl);
    if (longVar > STRING_LEN) longVar = STRING_LEN;
    strncpy(PECPLUGIN_pluginIndicationIDArray[arrayInd].optAndArgs, (char *)valstring->hdl, longVar);
    propertyMode = instance->ft->getProperty(instance, "Mode", &status);
    *mode = (functionalityMode_t)(propertyMode.value.uint16);
    if (status.msg) CMRelease(status.msg);
}

/*This function gets the Indication of a Functionality*/
static void getFuncIndication(const CMPIBroker *broker, 
                              const CMPIContext *ctx,
                              CMPIInstance *instance, 
                              pluginIndicationID_t *PEC_pluginIndicationIDArray, 
                              int arrayInd)
{
    CMPIStatus status;
    CMPIEnumeration *enumFuncIndAssoc;
    CMPIObjectPath *objectpath1 = instance->ft->getObjectPath(instance, NULL);
    CMPIObjectPath *opFuncIndAssoc;
    CMPIObjectPath *opFunctionality;
    CMPIData dataFuncIndAssoc;
    CMPIData propertyFunc;
    CMPIData propertyIndication;
    
    opFuncIndAssoc = broker->eft->newObjectPath(broker, NAMESPACE, FUNC_IND_ASSOCIATION, NULL);
    enumFuncIndAssoc = broker->bft->enumerateInstances(broker, ctx, opFuncIndAssoc, NULL, &status);
    if (!status.rc) 
    {
        while (enumFuncIndAssoc->ft->hasNext(enumFuncIndAssoc, NULL)) 
        {
            dataFuncIndAssoc = enumFuncIndAssoc->ft->getNext(enumFuncIndAssoc, NULL);
            propertyFunc = dataFuncIndAssoc.value.inst->ft->getProperty(dataFuncIndAssoc.value.inst, "Functionality", &status);
            opFunctionality = propertyFunc.value.ref;
            if (compareObjectPathString(objectpath1, opFunctionality, "InstanceID"))
            {
                propertyIndication = dataFuncIndAssoc.value.inst->ft->getProperty(dataFuncIndAssoc.value.inst, "Indication", &status);
                opFunctionality = propertyIndication.value.ref;
                getIndication(broker, ctx, opFunctionality, PEC_pluginIndicationIDArray, arrayInd);
            }
        }
    }
    else
    {
        debugOutput("Error in enumerate instances in PEC_FunctionalityIndicationSupport, rc=%d, msg=%s\n", 
               status.rc, (status.msg)? (char *)status.msg->hdl : NULL);      
    }
    if (enumFuncIndAssoc) CMRelease(enumFuncIndAssoc);
    if (opFuncIndAssoc) CMRelease(opFuncIndAssoc);
    if (status.msg) CMRelease(status.msg);

}

/*This function gets the Indications of an Object Path*/
static void getIndication(const CMPIBroker *broker, 
                          const CMPIContext *ctx,
                          CMPIObjectPath *objectpath1,
                          pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                          int arrayInd)
{
    CMPIStatus status;
    CMPIEnumeration *enumIndication;
    CMPIData keyIndication;
    char query[QUERY_FUNC_INDIC_LEN + 1];
    
    keyIndication = objectpath1->ft->getKey(objectpath1, "IndicationIdentifier", &status);
    int longVar = strlen((char *)((keyIndication.value.string)->hdl));
    if (longVar > INSTANCE_ID_LEN) debugOutput("Error in Indication Identified");
    snprintf(query, QUERY_FUNC_INDIC_LEN, SELECT_INDICATION, 
             PLUGIN_FUNC_INDICATION_CLASS, (char *)((keyIndication.value.string)->hdl));
    enumIndication = broker->bft->execQuery(broker, ctx, objectpath1, query, "WQL", &status);
    if (!status.rc)
    {
        while (enumIndication->ft->hasNext(enumIndication, NULL)) 
        {
            CMPIData dataIndication = enumIndication->ft->getNext(enumIndication, NULL);
            getIndData(dataIndication.value.inst, PECPLUGIN_pluginIndicationIDArray, arrayInd);
        }
    }
    else
    {
         debugOutput("Error in execQuery in PEC_RelatedIndication, rc=%d, msg=%s\n", 
                status.rc, (status.msg)? (char *)status.msg->hdl : NULL);      
    }
    if (enumIndication) CMRelease(enumIndication);
    if (status.msg) CMRelease(status.msg);
}

/*This function gets the data of an Indication*/
static void getIndData(CMPIInstance *instance, 
                       pluginIndicationID_t *PECPLUGIN_pluginIndicationIDArray, 
                       int arrayInd)
{
    CMPIStatus status;
    CMPIData propertyIndicationId;
    CMPIData propertyProbCause;
    CMPIData propertyFrec;
    CMPIString *valstring;

    CMPIObjectPath *opIndication = instance->ft->getObjectPath(instance, NULL);
    propertyIndicationId = instance->ft->getProperty(instance, "IndicationIdentifier", &status);
    valstring = propertyIndicationId.value.string;
    int longVar = strlen((char *)valstring->hdl);
    if (longVar > STRING_LEN) longVar = STRING_LEN;
    strncpy(PECPLUGIN_pluginIndicationIDArray[arrayInd].indicationIdentifier, (char *)valstring->hdl, longVar);
    if (status.msg) CMRelease(status.msg);
    propertyProbCause = instance->ft->getProperty(instance, "DefaultProbableCause", &status);
    valstring = propertyProbCause.value.string;
    longVar = strlen((char *)valstring->hdl);
    if (longVar > STRING_LEN) longVar = STRING_LEN;
    strncpy(PECPLUGIN_pluginIndicationIDArray[arrayInd].probableCause, (char *)valstring->hdl, longVar);
    if (status.msg) CMRelease(status.msg);
    propertyFrec = instance->ft->getProperty(instance, "Frequency", &status);
    PECPLUGIN_pluginIndicationIDArray[arrayInd].frequency = propertyFrec.value.uint16;
    if (status.msg) CMRelease(status.msg);
}
